from __future__ import annotations

import logging
import re
from urllib import parse

import markdown
import premailer
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.mail import EmailMultiAlternatives
from django.template import loader
from django.urls import reverse
from django.utils import translation
from django.utils.safestring import mark_safe

from emark import conf, utils

INLINE_LINK_RE = re.compile(r"\[[^\]]+\]\((https?://[^)]+)\)")
INLINE_HTML_LINK_RE = re.compile(r'href="https?://([^"]+)"')
CLS_NAME_TO_CAMPAIGN_RE = re.compile(
    r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


class MarkdownEmail(EmailMultiAlternatives):
    """Multipart email message that renders both plaintext and HTML from markdown.

    This is a full class:`EmailMultiAlternatives` subclass with additional support
    to send emails directly to a users. It also requires an explicit language
    (if not provided by the user) to ensure the correct language is used.

    Markdown templates are rendered using Django's template engine. The markdown
    may contain HTML tags. The HTML tags are rendered as-is, except for the plain
    text body, which extracts a Gmail style plain text version of the fully rendered
    HTML email.
    """

    base_html_template = "emark/base.html"
    template = None
    subject = None
    uuid = False

    def __init__(
        self,
        language: str,
        subject: str = "",
        context: dict = None,
        utm_params=None,
        template=None,
        **kwargs,
    ):
        """Initialize an email message and attach a rendered HTML version of it."""
        self.template = template or self.template
        self.context = context or {}
        self.language = language
        self.utm_params = utm_params or {}
        self.subject = subject or self.subject
        self.html = None
        self.markdown = None
        super().__init__(subject=self.subject, **kwargs)

    @classmethod
    def to_user(cls, user=None, context=None, language=None, **kwargs):
        """Return email with user specific language, context and recipient."""
        if not user.email:
            raise ValueError("User has no email address")
        if not user.is_active:
            raise ValueError("User is not active")
        if settings.USE_I18N:
            try:
                language = language or user.language
            except AttributeError as e:
                raise ValueError(
                    "If your user model does not have a language field,"
                    " you must provide a language."
                ) from e
        context = context or {}
        context |= {
            "short_name": user.get_short_name(),
            "user": user,
        }
        obj = cls(
            **{
                "to": [f'"{user.get_full_name()}" <{user.email}>'],
                "context": context,
                "language": language,
            }
            | kwargs
        )
        obj.user = user
        return obj

    def message(self):
        # The connection will call .message while sending the email.
        self.render()
        return super().message()

    @classmethod
    def get_utm_campaign_name(cls):
        """Return the UTM campaign name for this email."""
        return "_".join(
            m.group(0) for m in CLS_NAME_TO_CAMPAIGN_RE.finditer(cls.__qualname__)
        ).upper()

    def update_url_params(self, url, **params):
        """Add UTM parameters to a URL and add the click tracking URL."""
        redirect_url_parts = parse.urlparse(url)
        url_params = dict(parse.parse_qsl(redirect_url_parts.query))
        params.update(url_params)
        url_new_query = parse.urlencode(params)
        redirect_url_parts = redirect_url_parts._replace(query=url_new_query)
        redirect_url = parse.urlunparse(redirect_url_parts)
        if not self.uuid:
            return redirect_url
        site_url = self.get_site_url()
        # external links should not be tracked
        top_level_domain = ".".join(site_url.split(".")[-2:])
        if not redirect_url_parts.netloc.endswith(top_level_domain):
            return redirect_url
        tracking_url = reverse("emark:email-click", kwargs={"pk": self.uuid})
        tracking_url = parse.urljoin(site_url, tracking_url)
        tracking_url_parts = parse.urlparse(tracking_url)
        tracking_url_parts = tracking_url_parts._replace(
            query=parse.urlencode({"url": redirect_url})
        )
        return parse.urlunparse(tracking_url_parts)

    def inject_utm_params(self, md, **utm):
        for url in INLINE_LINK_RE.findall(md):
            md = md.replace(f"({url})", f"({self.update_url_params(url, **utm)})")
        for url in INLINE_HTML_LINK_RE.findall(md):
            md = md.replace(
                f'href="{url}"', f'href="{self.update_url_params(url, **utm)}"'
            )
        return md

    def get_template(self):
        if not self.template:
            raise ImproperlyConfigured(
                f"{self.__class__.__qualname__} is missing a template."
            )
        return self.template

    def get_site_url(self):
        protocol = "https" if settings.SECURE_SSL_REDIRECT else "http"
        if domain := conf.get_settings().DOMAIN:
            pass
        elif apps.is_installed("django.contrib.sites"):
            from django.contrib.sites.models import Site

            domain = Site.objects.get_current().domain

        return parse.urlunparse((protocol, domain, "", "", "", ""))

    def get_utm_params(self) -> {str: str}:
        """Return a dictionary of UTM parameters."""
        return (
            conf.get_settings().UTM_PARAMS
            | {
                "utm_campaign": self.get_utm_campaign_name(),
            }
            | self.utm_params
        )

    def get_context_data(self):
        """Return the context data for the email."""
        context = {}
        if self.uuid:
            context |= {
                "tracking_uuid": self.uuid,
                "view_in_browser_url": parse.urljoin(
                    self.get_site_url(),
                    reverse("emark:email-detail", kwargs={"pk": self.uuid}),
                ),
                "tracking_pixel_url": parse.urljoin(
                    self.get_site_url(),
                    reverse("emark:email-open", kwargs={"pk": self.uuid}),
                ),
            }

        return context | self.context

    def get_subject(self, **context):
        """Return the email's subject."""
        if not self.subject:
            raise ImproperlyConfigured(
                f"{self.__class__.__qualname__} is missing a subject."
            )
        return self.subject % context

    def get_preheader(self):
        """Return the email's preheader.

        A brief text that recipients will see in their inbox before opening the email
        along with the subject. Unless explicitly set, the preheader will be the first
        paragraph of the email's body.
        Can be useful to grab the recipient's attention and/or simplify their workflow
        (e.g. by providing a one-time-password).
        """
        return ""

    def get_markdown(self, context, utm):
        template = self.get_template()
        markdown_string = loader.get_template(template).render(context)
        return self.inject_utm_params(markdown_string, **utm)

    def get_html(self, markdown_string, context):
        html_message = markdown.markdown(
            markdown_string,
            extensions=[
                "markdown.extensions.meta",
                "markdown.extensions.tables",
                "markdown.extensions.extra",
            ],
        )
        context["markdown_string"] = mark_safe(html_message)  # noqa: S308

        template = loader.get_template(self.base_html_template)
        rendered_html = template.render(context)

        inlined_html = premailer.transform(
            html=rendered_html,
            strip_important=False,
            keep_style_tags=True,
            cssutils_logging_level=logging.ERROR,
        )
        return inlined_html

    def get_body(self, html):
        """Return the parsed plain text version of the rendered HTML email."""
        parser = utils.HTML2TextParser()
        parser.feed(html)
        parser.close()
        return str(parser)

    def render(self, tracking_uuid=None):
        """Render the email."""
        if self.html is None:
            self.uuid = tracking_uuid
            with translation.override(self.language):
                utm_params = self.get_utm_params()
                context = self.get_context_data()
                context |= utm_params
                self.subject = self.get_subject(**context)
                context["subject"] = self.subject
                context["preheader"] = self.get_preheader()
                self.markdown = self.get_markdown(context, utm_params)
                self.html = self.get_html(
                    markdown_string=self.markdown,
                    context=context,
                )
                self.body = self.get_body(self.html)
                self.attach_alternative(self.html, "text/html")



import uuid

from django.conf import settings
from django.db import models
from django.urls import reverse


class Send(models.Model):
    """Frozen replica of a sent email message."""

    uuid = models.UUIDField(
        unique=True, default=uuid.uuid4, editable=False, primary_key=True
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name="emark_emails",
        null=True,
    )
    # In RFC 2822 "from" is a mailbox-list, but Django only support a single
    from_email = models.TextField(max_length=998)
    to = models.JSONField(default=list)
    cc = models.JSONField(default=list)
    bcc = models.JSONField(default=list)
    reply_to = models.JSONField(default=list)
    subject = models.TextField(max_length=998)  # RFC 2822
    body = models.TextField()
    html = models.TextField(null=True)
    utm = models.JSONField(default=dict)
    created_at = models.DateTimeField(auto_now_add=True)

    def get_absolute_url(self):
        return reverse("emark:email-detail", kwargs={"pk": self.pk})


class ClientTrackingQueryset(models.QuerySet):
    def create_for_request(self, request, **kwargs):
        """Create a tracking record for the given request."""
        return self.create(
            headers=dict(request.headers),
            ip_address=request.META.get("REMOTE_ADDR"),
            utm={
                key: value
                for key, value in request.GET.items()
                if key.startswith("utm_")
            },
            **kwargs,
        )


class ClientTrackingModelMixin(models.Model):
    uuid = models.UUIDField(
        unique=True, default=uuid.uuid4, editable=False, primary_key=True
    )
    headers = models.JSONField(default=dict)
    ip_address = models.GenericIPAddressField()
    utm = models.JSONField(default=dict)

    objects = ClientTrackingQueryset.as_manager()

    class Meta:
        abstract = True


class Click(ClientTrackingModelMixin):
    """Record of a click on a link in an email."""

    email = models.ForeignKey(Send, on_delete=models.CASCADE)
    # we don't need validation here, but we do need to store long URLs
    redirect_url = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)


class Open(ClientTrackingModelMixin):
    """Record of an email being opened."""

    email = models.ForeignKey(Send, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)



import uuid

from django.core.mail import EmailMessage
from django.core.mail.backends.console import EmailBackend as _ConsoleEmailBackend
from django.core.mail.backends.smtp import EmailBackend as _SMTPEmailBackend

from emark import models
from emark.message import MarkdownEmail

__all__ = [
    "ConsoleEmailBackend",
    "SMTPEmailBackend",
    "TrackingConsoleEmailBackend",
    "TrackingSMTPEmailBackend",
]


class RenderEmailBackendMixin:
    def send_messages(self, email_messages):
        for message in email_messages:
            if isinstance(message, MarkdownEmail):
                message.render()
        return super().send_messages(email_messages)


class TrackingEmailBackendMixin:
    """Add a tracking framework to an email backend."""

    def send_messages(self, email_messages):
        self._messages_sent = []
        for message in email_messages:
            if isinstance(message, MarkdownEmail):
                message.render(tracking_uuid=uuid.uuid4())
        try:
            return super().send_messages(email_messages)
        finally:
            models.Send.objects.bulk_create(self._messages_sent)

    def _track_message(self, message: EmailMessage):
        if isinstance(message, MarkdownEmail):
            self._messages_sent.append(
                models.Send(
                    pk=message.uuid,
                    from_email=message.from_email,
                    to=message.to,
                    cc=message.cc,
                    bcc=message.bcc,
                    reply_to=message.reply_to,
                    subject=message.subject,
                    body=message.body,
                    html=message.html,
                    user=getattr(message, "user", None),
                    utm=message.get_utm_params(),
                )
            )
        else:
            self._messages_sent.append(
                models.Send(
                    from_email=message.from_email,
                    to=message.to,
                    cc=message.cc,
                    bcc=message.bcc,
                    reply_to=message.reply_to,
                    subject=message.subject,
                    body=message.body,
                )
            )


class ConsoleEmailBackendMixin:
    """Drop email alternative parts and attachments for the console backend."""

    def write_message(self, message):
        msg = message.message()
        if not msg.is_multipart():
            return super().write_message(message)

        original_payload = msg.get_payload()
        payload_count = len(original_payload)
        msg.set_payload(msg.get_payload(0))
        msg_data = msg.as_bytes()
        charset = (
            msg.get_charset().get_output_charset() if msg.get_charset() else "utf-8"
        )
        msg_data = msg_data.decode(charset)
        self.stream.write(f"{msg_data}\n")
        self.stream.write("-" * 79)
        self.stream.write("\n")
        if payload_count > 1:
            self.stream.write(
                f"{payload_count - 1} more attachment(s) have been omitted.\n"
            )

        return msg


class ConsoleEmailBackend(
    RenderEmailBackendMixin, ConsoleEmailBackendMixin, _ConsoleEmailBackend
):
    """Like the console email backend but only with the plain text body."""


class SMTPEmailBackend(RenderEmailBackendMixin, _SMTPEmailBackend):
    """SMTP email backend that renders messages before establishing an SMTP transport."""

    pass


class TrackingConsoleEmailBackend(
    TrackingEmailBackendMixin, ConsoleEmailBackendMixin, _ConsoleEmailBackend
):
    """Like the console email backend but with click and open tracking."""

    def write_message(self, message):
        try:
            return super().write_message(message)
        finally:
            self._track_message(message)


class TrackingSMTPEmailBackend(TrackingEmailBackendMixin, _SMTPEmailBackend):
    """Like the SMTP email backend but with click and open tracking.

    Furthermore, all emails are sent to a single email address.
    If multiple to, cc, or bcc addresses are specified, a separate
    email is sent individually to each address.
    """

    def _send(self, email_message):
        sent = False
        try:
            sent = super()._send(email_message)
            return sent
        finally:
            if sent:
                self._track_message(email_message)



import json

from django.db import migrations, models


def forwards_func(apps, schema_editor):  # pragma: no cover
    Send = apps.get_model("emark", "Send")
    for send in Send.objects.all():
        send.to = json.dumps([send.to])
        send.save(update_fields=["to"])


class Migration(migrations.Migration):
    dependencies = [
        ("emark", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="send",
            old_name="from_address",
            new_name="from_email",
        ),
        migrations.RenameField(
            model_name="send",
            old_name="to_address",
            new_name="to",
        ),
        migrations.RunPython(
            forwards_func,
        ),
        migrations.AlterField(
            model_name="send",
            name="to",
            field=models.JSONField(default=list),
        ),
        migrations.AddField(
            model_name="send",
            name="cc",
            field=models.JSONField(default=list),
        ),
        migrations.AddField(
            model_name="send",
            name="bcc",
            field=models.JSONField(default=list),
        ),
        migrations.AddField(
            model_name="send",
            name="reply_to",
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name="send",
            name="from_email",
            field=models.TextField(max_length=998),
        ),
    ]



from django.urls import path

from . import views

app_name = "emark"
urlpatterns = [
    path("<uuid:pk>/", views.EmailDetailView.as_view(), name="email-detail"),
    path("<uuid:pk>/click", views.EmailClickView.as_view(), name="email-click"),
    path("<uuid:pk>/open", views.EmailOpenView.as_view(), name="email-open"),
]



# Generated by Django 4.2.2 on 2023-06-25 09:50

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Send",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="emark_emails",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("from_address", models.EmailField(max_length=254)),
                ("to_address", models.EmailField(max_length=254)),
                ("subject", models.TextField(max_length=998)),
                ("body", models.TextField()),
                ("html", models.TextField(null=True)),
                ("utm", models.JSONField(default=dict)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="Open",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("headers", models.JSONField(default=dict)),
                ("ip_address", models.GenericIPAddressField()),
                ("utm", models.JSONField(default=dict)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "email",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="emark.send"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Click",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("headers", models.JSONField(default=dict)),
                ("ip_address", models.GenericIPAddressField()),
                ("utm", models.JSONField(default=dict)),
                ("redirect_url", models.TextField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "email",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="emark.send"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from urllib.parse import urlparse

from django import http
from django.conf import settings
from django.http.request import split_domain_port, validate_host
from django.views import View
from django.views.generic.detail import SingleObjectMixin

from . import models

# white 1x1 pixel JPEG in bytes:
#
# import io
# from PIL import Image
#
# img = Image.new('RGB', (1, 1), color='white')
# img_bytes = io.BytesIO()
# img.save(img_bytes, format='GIF')
# TRACKING_PIXEL_GIF = img_bytes.getvalue()
TRACKING_PIXEL_GIF = b"GIF87a\x01\x00\x01\x00\x81\x00\x00\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x08\x04\x00\x01\x04\x04\x00;"


class EmailDetailView(SingleObjectMixin, View):
    """Return the HTML body of the email."""

    model = models.Send

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        if self.object.html:
            return http.HttpResponse(
                self.object.html.encode(), status=200, content_type="text/html"
            )
        return http.HttpResponse(
            self.object.body.encode(), status=200, content_type="text/plain"
        )


class EmailClickView(SingleObjectMixin, View):
    """Redirect to the URL and track the click."""

    model = models.Send

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        redirect_to = request.GET.get("url")
        # The redirect_to URL is user-provided, so it might be malicious
        # or malformed. We use Django's URL validation to ensure that it
        # is safe to redirect to.
        parsed_url = urlparse(redirect_to)
        if not parsed_url.netloc:
            return http.HttpResponseBadRequest("Missing url or malformed parameter")

        domain, _port = split_domain_port(parsed_url.netloc)
        allowed_hosts = settings.ALLOWED_HOSTS
        if settings.DEBUG:
            allowed_hosts = settings.ALLOWED_HOSTS + [
                ".localhost",
                "127.0.0.1",
                "[::1]",
            ]
        if any(
            [
                not domain,
                not validate_host(domain, allowed_hosts),
                request.scheme != parsed_url.scheme,
            ]
        ):
            return http.HttpResponseBadRequest("Missing url or malformed parameter")

        models.Click.objects.create_for_request(
            request, email=self.object, redirect_url=redirect_to
        )
        return http.HttpResponseRedirect(redirect_to)


class EmailOpenView(SingleObjectMixin, View):
    """Return a tracking pixel and track the open."""

    model = models.Send

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        models.Open.objects.create_for_request(request, email=self.object)

        return http.HttpResponse(
            TRACKING_PIXEL_GIF,
            status=200,
            content_type="image/gif",
            headers={
                "Cache-Control": "no-cache, no-store, must-revalidate",
            },
        )
